/*
 * the dvgrab main program
 * contains the main logic
 */

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <assert.h> 
#include <fcntl.h> 
#include <unistd.h> 
#include <stdio.h> 
#include <stdlib.h> 
#include <signal.h> 
#include <pthread.h> 
#include <string.h>

#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <unistd.h>

#include "rtp_api.h"
#include "rtp_highlevel.h"
  
#define tv2dbl(tv) ((tv).tv_sec + (tv).tv_usec / 1000000.0)

#define RAW_BUF_SIZE    (10240)

enum { PAL_FORMAT, NTSC_FORMAT, AVI_DV1_FORMAT, AVI_DV2_FORMAT, QT_FORMAT, RAW_FORMAT, TEST_FORMAT, UNDEFINED };


int    g_buffer_underrun;
volatile int    g_reader_active;
volatile int     g_alldone;
int     g_filedone;
char    *g_dst_file_name;
int     g_frame_size;

int     g_autosplit;
int     g_timestamp;
int     g_card;
int     g_channel;
int     g_file_format;
int     g_index_type;
int     g_frame_count;
int     g_frame_every;
int     g_frame_to_skip;
int     g_progress;
int     g_testmode;

#ifdef FREEBSD_5
int device;
#endif /* FREEBSD_5 */

/*
pthread_mutex_t g_mutex;
pthread_t       g_thread;
*/

char addr[20];
int port = 12000;
 
context cid;

int dump_fd;
int flags_dump = 0;
int flags_file = 0;
int flags_firewire = 1;

char *filename = NULL;

unsigned char ttl = 255;
rtperror status;

struct evt_queue_elt {
  context cid;
  rtp_opaque_t event_opaque;
  double event_time;
  struct evt_queue_elt *next;
};

static struct evt_queue_elt* evt_queue = NULL; 

static void insert_in_evt_queue(struct evt_queue_elt *elt)
{
  if (evt_queue == NULL || elt->event_time < evt_queue->event_time) {
    elt->next = evt_queue;
    evt_queue = elt;
  }
  else {
    struct evt_queue_elt *s = evt_queue;
    while (s != NULL) {
      if (s->next == NULL || elt->event_time < s->next->event_time) {
        elt->next = s->next;
        s->next = elt;
        break;
      }
      s = s->next;
    }
  }
}

void RTPSchedule(context cid, rtp_opaque_t opaque, struct timeval *tp)
{
  struct evt_queue_elt *elt;

  elt = (struct evt_queue_elt *) malloc(sizeof(struct evt_queue_elt));
  if (elt == NULL)
    return;

  elt->cid = cid;
  elt->event_opaque = opaque;
  elt->event_time = tv2dbl(*tp);

  insert_in_evt_queue(elt);

  return;
}


void getargs(int argc, char *argv[])
{
  int i;

  g_autosplit = 0;
  g_timestamp = 0;
  g_card = 0;
  g_channel = 63;
  g_frame_count = 100000;
  g_frame_every = 1;
  g_dst_file_name = NULL;
  g_progress = 0;
  g_testmode = 0;

  for (i = 1; i < argc; ++i) {
    if (strcmp("--autosplit", argv[i]) == 0) {
      g_autosplit = 1;
    } else if ((strcmp("--format", argv[i]) == 0) && (i < argc)) {
      i++;
      if (strcmp("raw", argv[i]) == 0) {
        g_file_format = RAW_FORMAT;
      } else {
        exit(1);
      }
    } else if ((strcmp("--card", argv[i]) == 0) && (i < argc)) {
      i++;
      if (sscanf(argv[i], "%d", &g_card) != 1) {
        exit(1);
      }
    } else if ((strcmp("--channel", argv[i]) == 0) && (i < argc)) {
      i++;
      if (sscanf(argv[i], "%d", &g_channel) != 1) {
          exit(1);
      }
    } else if ((strcmp("--ip", argv[i]) == 0) && (i < argc)) {
      i++;
      strncpy(addr, argv[i], 20);
    } else if ((strcmp("--port", argv[i]) == 0) && (i < argc)) {
      i++;
      if (sscanf(argv[i], "%d", &port) != 1) {
        exit(1);
      }
    } else if ((strcmp("--file", argv[i]) == 0) && (i <argc)) {
      i++;
      flags_file = 1;
      flags_firewire = 0;
      filename = argv[i];
    } else if ((strcmp("--dump", argv[i]) == 0) && (i < argc)) {
      flags_dump = 1;
    } else if (strcmp("--help", argv[i]) == 0) {
        printf("Usage: %s [options]\n", argv[0]);
        printf("\nOptions:\n\n");
        printf("  --autosplit       start a new file when the frame count is exceeded, a new\n");
        printf("                    recording is detected, or file becomes too big.\n");
        printf("                    Note: will loop forever -- you need to press ctrl-c\n");
        printf("                    at the end.\n");
        printf("                    Default is autosplit when file is slightly less than\n");
        printf("                    1 GByte.\n");
        printf("  --format raw      save only the raw DV information.\n");
	printf("  --dump            dump dv data to file.\n");
        printf("  --card number     card number (default %d).\n", g_card);
        printf("  --channel number  iso channel number for listening (default %d).\n", g_channel);
        printf("                    instead from camcorder.\n");
	printf("  --ip              dest. ip address.\n");
	printf("  --port            dest. ip port.\n");
	printf("  --file	    read form file instead from firewire.\n");
        exit(0);
    } else {
        exit(1);
    }
  }
  if (addr == NULL) {
      exit(1);
  }
}

#ifdef LINUX
extern int capture_raw_linux();
#endif

#ifdef FREEBSD_5
extern int capture_raw_freebsd(int d, char ich);
#endif

extern int capture_raw_file(char *filename);

void signal_handler(int sig)
{
  /* replace this signal handler with the default (which aborts) */

  signal(SIGINT, SIG_IGN);

  /* setting these variables will let us fall out of the main loop */

  g_reader_active = 0;
  g_alldone = 1;

#ifdef FREEBSD_5
  close(device);
#endif
  
  status = RTPCloseConnection(cid, "Goodbye!");
  if (status != RTP_OK){
    printf("RTPCloseConnection: %s\n", RTPStrError(status));
  }
  status = RTPDestroy(cid);
  if (status != RTP_OK){
    printf("RTPDestroy: %s\n", RTPStrError(status));
  }

}


int main(int argc, char *argv[])
{
#ifdef FREEBSD_5
  char devname[] = "/dev/fw0";
#endif

  getargs(argc, argv);

  signal(SIGINT, signal_handler);

  printf("Using host %s/%d\n",addr, port);

  status = RTPCreate(&cid);
  if (status != RTP_OK){
      fprintf(stderr, "RTPCreate: %s\n", RTPStrError(status));
  }

  status = RTPSessionAddSendAddr(cid, addr, port, ttl);
  if (status != RTP_OK){
    printf("RTPSessionAddSendAddr: %s\n", RTPStrError(status));
  }

  status = RTPSessionSetReceiveAddr(cid, addr, port);
  if (status != RTP_OK){
    printf("RTPSessionSetReceiveAddr: %s\n", RTPStrError(status));
  }

  status = RTPOpenConnection(cid);
  if (status != RTP_OK){
    printf("RTPOpenConnection: %s\n", RTPStrError(status));
  }

  if (flags_dump) {
#ifdef __USE_LARGEFILES64
    dump_fd = open("dump.dv", O_RDWR | O_CREAT | O_LARGEFILE, 00644);
#else
    dump_fd = open("dump.dv", O_RDWR | O_CREAT, 00644);
#endif /* __USE_LARGEFILES64 */

    if (dump_fd < 0) {
      perror("open dump.dv");
      flags_dump = 0;
    }
  }		  

if (flags_firewire){
#ifdef LINUX
  capture_raw_linux();
#endif
  
#ifdef FREEBSD_5
  device = open(devname, O_RDWR);
  if (device < 0) {
    perror("open");

    exit(1);
  }
  
  capture_raw_freebsd(device, (1<<6) | 63);
#endif
} else {
  capture_raw_file(filename);
}
      
  return 0;
}

