#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include <unistd.h>
#include <fcntl.h>

#ifdef LINUX
#include <assert.h>
#include <signal.h>
#include <pthread.h>

#include <libraw1394/raw1394.h>
#include <libraw1394/csr.h>
#endif /* LINUX */

#ifdef FREEBSD_5
#include <err.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <errno.h>
#include <sys/time.h>  /* struct timeval, gettimeofday() */
#include <netdb.h>
#include <unistd.h>

#include <sys/param.h>
#include <sys/ioctl.h>
#include <sys/types.h>
#include <sys/uio.h>

#if __FreeBSD_version >= 500000
#include <arpa/inet.h>
#endif

#include <dev/firewire/firewire.h>
#include "iec68113.h"
#endif /* FREEBSD_5 */

#include "dvsend.h"
#include "rtp.h"

#include "rtp_api.h"
#include "rtp_highlevel.h"
  
#define RAW_BUF_SIZE	(10240)

/* global variables shared with the main thread */

#ifdef FREEBSD_5
#define DEBUG		0
#define FIX_FRAME	1

#define PSIZE 512
#define DSIZE 480
#define NCHUNK 8

#define NPACKET_R 256
#define NPACKET_T 255

#define NPKT 50
#define NVEC 50
#define BLOCKSIZE 80

#define TNBUF 100       /* XXX too large value causes block noise */
#define NEMPTY 10       /* depends on TNBUF */

#define RBUFSIZE (PSIZE * NPACKET_R)
#define MAXBLOCKS (300)
#define CYCLE_FRAC 0xc00
#endif /* FREEBSD_5 */

#ifdef FREEBSD_5
char *system_name[] = {"NTSC", "PAL"};
int frame_rate[] = {30, 25};

unsigned long int SentPackets = 0;
unsigned long int dvdata = 0;
#endif /* FREEBSD_5 */

#ifdef LINUX
int	g_testdata;
unsigned char *g_frame;

char *g_dst_file_name;
int g_frame_count;

volatile int	g_reader_active;
int	g_buffer_underrun;
int	g_card;
int	g_channel;

int raw_iso_handler(raw1394handle_t handle, int channel, size_t length,
                    quadlet_t *data)
{
        if (length < RAW_BUF_SIZE) {
                *(int*)g_frame = length;
                memcpy(g_frame + 4, data, length);
        }
        return 0;
}


int my_reset_handler(raw1394handle_t handle, unsigned int generation)
{
        static int i = 0;

        printf("reset %d\n", i++);
        if (i == 100)
                g_reader_active = 0;
        return 0;
}

raw1394handle_t open_1394_driver(int channel, iso_handler_t handler)
{
        int numcards;
        struct raw1394_portinfo g_pinf[16];
        iso_handler_t	g_oldhandler;
        raw1394handle_t handle;

        if (!(handle = raw1394_new_handle())) {
                perror("raw1394 - couldn't get handle");
		printf("This error usually means that the ieee1394 driver is not loaded or that /dev/raw1394 does not exist.\n");
                exit( -1);
        }

        if ((numcards = raw1394_get_port_info(handle, g_pinf, 16)) < 0) {
                perror("raw1394 - couldn't get card info");
                exit( -1);
        }

        if (raw1394_set_port(handle, g_card) < 0) {
                perror("raw1394 - couldn't set port");
                exit( -1);
        }

        g_oldhandler = raw1394_set_iso_handler(handle, g_channel, handler);
        /*      raw1394_set_tag_handler(handle, my_tag_handler);              */
        raw1394_set_bus_reset_handler(handle, my_reset_handler);

        /* Starting iso receive */

        if (raw1394_start_iso_rcv(handle, channel) < 0) {
                perror("raw1394 - couldn't start iso receive");
                exit( -1);
        }
        return handle;
}


void close_1394_driver(int channel, raw1394handle_t handle)
{
	raw1394_stop_iso_rcv(handle, channel);
        raw1394_destroy_handle(handle);
}

int capture_raw_linux()
{
  int frames_read;
  int length;
  raw1394handle_t handle;
  int found_first_frame;
  int skipped = 0;
  int	g_alldone=0;

  g_frame = (unsigned char*)malloc(RAW_BUF_SIZE);

  handle = open_1394_driver(g_channel, raw_iso_handler);

  frames_read = 0;
  found_first_frame = 0;

  while (!g_alldone) {
    raw1394_loop_iterate(handle);
    length = *(int*)g_frame;
    if (length >= 492) {
      if (!found_first_frame) {
        if (g_frame[16] == 0x1f && g_frame[17] == 0x07)
          found_first_frame = 1;
        else skipped++;
      }
      if (skipped > 500) {
        printf("skipped too many without finding frame\n");
        break;
      }
      if (found_first_frame) {
          if (g_frame[16] == 0x1f && g_frame[17] == 0x07)
            frames_read++;
          if (frames_read > g_frame_count)
            break;
	  if (flags_dump) {
          write(dump_fd, (g_frame + 16), 480);
	  }
	  dvsend_RTPSend((int8*)(g_frame + 16), 480);
      }
    }
  }
  close_1394_driver(g_channel, handle);
  free(g_frame);

  return 0;
}
#endif /* LINUX */


#ifdef FREEBSD_5
int capture_raw_freebsd(int d, char ich)
{
	struct fw_isochreq isoreq;
	struct fw_isobufreq bufreq;
	struct dvdbc *dv;
	struct ciphdr *ciph;
	struct fw_pkt *pkt;
	char *pad, *buf;
	u_int32_t *ptr;
	int len, tlen, npad, k, m, vec, nb, system = -1 ;
	int nblocks[] = {250 /* NTSC */, 300 /* PAL */};
	struct iovec wbuf[NPACKET_R];

        int pad_iter, iovec_iter;

	int	g_alldone = 0;
	

#if 0
	int mode;
#endif

	buf = (char *)malloc(RBUFSIZE);
	pad = (char *)malloc(DSIZE*MAXBLOCKS);
	
	/*bzero(pad, BLOCKSIZE*MAXBLOCKS);*/
	memset(pad, 0xff, DSIZE*MAXBLOCKS);
	bzero(wbuf, sizeof(wbuf));
	
#if 0	
	if (ioctl(d, FW_SSTDV, &mode) < 0)
       		err(1, "ioctl");
#else
	bufreq.rx.nchunk = NCHUNK;
	bufreq.rx.npacket = NPACKET_R;
	bufreq.rx.psize = PSIZE;
	bufreq.tx.nchunk = 0;
	bufreq.tx.npacket = 0;
	bufreq.tx.psize = 0;
	if (ioctl(d, FW_SSTBUF, &bufreq) < 0) {
		err(1, "%s", "ioctl");
	}
#endif

	isoreq.ch = ich & 0x3f;
	isoreq.tag = (ich >> 6) & 3;

	if( ioctl(d, FW_SRSTREAM, &isoreq) < 0)
       		err(1, "%s", "ioctl");

	k = m = 0;
	while (!g_alldone) {
#if 0
                tlen = 0;
                while ((len = read(d, buf + tlen, PSIZE
                                                /* RBUFSIZE - tlen */)) > 0) {
                        if (len < 0) {
                                if (errno == EAGAIN) {
                                        fprintf(stderr, "(EAGAIN)\n");
                                        fflush(stderr);
                                        if (len <= 0)
                                                continue;
                                } else
                                        err(1, "read failed");
                        }
                        tlen += len;
                        if ((RBUFSIZE - tlen) < PSIZE)
                                break;
                };
#else
                tlen = len = read(d, buf, RBUFSIZE);
                if (len < 0) {
                        if (errno == EAGAIN) {
                                fprintf(stderr, "(EAGAIN)\n");
                                fflush(stderr);
                                if (len <= 0)
                                        continue;
                        } else
                                err(1, "read failed");
                }
#endif
	
		vec = 0;
		/*wbuf[0].iov_len = 0;*/
		ptr = (u_int32_t *) buf;
again:
		pkt = (struct fw_pkt *) ptr;

#if DEBUG
		printf("%08x %08x %08x %08x\n",
			htonl(ptr[0]), htonl(ptr[1]),
			htonl(ptr[2]), htonl(ptr[3]));
#endif /* DEBUG */

		ciph = (struct ciphdr *)(ptr + 1);	/* skip iso header */
                if (ciph->fmt != CIP_FMT_DVCR)
                        errx(1, "unknown format 0x%x", ciph->fmt);
		ptr = (u_int32_t *) (ciph + 1);		/* skip cip header */

#if DEBUG
                if (ciph->fdf.dv.cyc != 0xffff && k == 0) { 
                        fprintf(stderr, "0x%04x\n", ntohs(ciph->fdf.dv.cyc));
                }
#endif /* DEBUG */

		if (pkt->mode.stream.len <= sizeof(struct ciphdr)) {
                        /* no payload */
			/*
  			  ptr ++; 	 
			  ptr += ntohs(pkt->mode.stream.len)/sizeof(u_int32_t);
			*/
			goto next;
		}
		
		for (dv = (struct dvdbc *)ptr;
				(char *)dv < (char *)(ptr + ciph->len);
				dv+=6) {

#if DEBUG
			printf("(%d,%d) ", dv->sct, dv->dseq);
#endif /* DEBUG */
			if  (dv->sct == DV_SCT_HEADER && dv->dseq == 0) {
                                if (system < 0) {
                                        system = ciph->fdf.dv.fs;
                                        printf("%s\n", system_name[system]);
                                }

                                /* Fix DSF bit */
                                if (system == 1 &&
                                        (dv->payload[0] & DV_DSF_12) == 0)
                                        dv->payload[0] |= DV_DSF_12;
                                nb = nblocks[system];
                                fprintf(stderr, "%d", k%10);

#if FIX_FRAME
				if (m > 0 && m != nb) {
					/* padding bad frame */
					npad = ((nb - m) % nb);
					if (npad < 0)
						npad += nb;
					printf("(%d blocks padded)", npad);
					npad *= DSIZE;
					wbuf[vec].iov_base = pad;
					wbuf[vec++].iov_len = npad;

					if (flags_dump) {
					  write(dump_fd, pad, npad);
					}
					if (vec >= NPACKET_R) {

                                        	printf("Packet length: %d\n", npad*BLOCKSIZE);
                                        	for(pad_iter=0; pad_iter < npad; pad_iter++){

						  dvsend_RTPSend(pad, BLOCKSIZE);
                                        	}
						
						vec = 0;
					}
				}
#endif /* FIX_FRAME */
				k++;
				if (k % frame_rate[system] == 0) { /* every second */
					printf("\n");
					fflush(stdout);
				}
				m = 0;
			}
			
			m++;
                        wbuf[vec].iov_base = (char *) dv;
                        wbuf[vec++].iov_len = DSIZE;

                        if (vec >= NPACKET_R) {
		  	        if (flags_dump) {
			          writev(dump_fd, wbuf, vec);
			        }
				for(iovec_iter=0; iovec_iter < vec; iovec_iter++){

				  	dvsend_RTPSend(wbuf[iovec_iter].iov_base, wbuf[iovec_iter].iov_len);
                                }

				vec = 0;
			}

		}
		ptr = (u_int32_t *)dv;
next:
		if ((char *)ptr < buf + tlen) {
			goto again;
		}

		if (vec > 0) {
		  	if (flags_dump) {
			  writev(dump_fd, wbuf, vec+1);
			}
                        for(iovec_iter=0; iovec_iter < vec+1; iovec_iter++){

				dvsend_RTPSend(wbuf[iovec_iter].iov_base, wbuf[iovec_iter].iov_len);

                        }

		}
	}
	printf("\n");
	return 0;
}
#endif /* FREEBSD_5 */
